import re
from collections import defaultdict

from cnki2bibtex.entry import Entry
from cnki2bibtex.misc.check import check_entry_has_valid_fields


class CNKIEntry(Entry):
    """
    One entry of CNKI .net file.
    """

    def __init__(self):
        self.fields = defaultdict(lambda: None)
        self.recorded_field_names = []

    def add_field_from_line(self, string_line):
        re_obj = re.match(r"\{(.*)\}:(.*)", string_line)
        field_name = re_obj.group(1).strip()
        field_content = re_obj.group(2).strip()
        self.fields[field_name] = field_content

    def add_all_fields_from_block(self, string_block):
        for line in string_block.split("\n"):
            self.add_field_from_line(line.strip())

    def __mark_field_is_recorded(self, key):
        if key not in self.fields:
            raise FieldNotInEntryException(key, self)
        elif key not in self.recorded_field_names:
            self.recorded_field_names.append(key)
        else:
            raise FieldAlreadyRecordedException(key, self)

    @check_entry_has_valid_fields
    def mark_fields_are_recorded(self, *keys):
        for key in keys:
            self.__mark_field_is_recorded(key)

    @check_entry_has_valid_fields
    def field_is_not_recorded(self, key):
        return key not in self.recorded_field_names


class CNKIEntryFactory(object):
    def give_all_entries(self, full_text_in_net_file):
        entries = []
        for block in full_text_in_net_file.strip().split(r"{Reference Type}")[1:]:
            block = r"{Reference Type}" + block.strip()
            block = self.preprocess(block)
            tmp = CNKIEntry()
            tmp.add_all_fields_from_block(block.strip())
            entries.append(tmp)
        return entries

    def preprocess(self, block):
        block = self.fix_line_break_in_content(block)
        return block

    def fix_line_break_in_content(self, block):
        block = re.sub("\n+", "\n", block)
        block = re.sub("\n([^\{])", lambda match_obj: match_obj.group(1), block)
        return block


class FieldNotInEntryException(Exception):
    def __init__(self, required_field, entry):
        if "Title" not in entry:
            self.message = "{} is not in the entry.".format(required_field)
        else:
            self.message = "{} is not in the entry titled {}.".format(required_field, entry["Title"])

        self.args = (self.message,)


class FieldAlreadyRecordedException(Exception):
    def __init__(self, marked_field, entry):
        if "Title" not in entry:
            self.message = "Field {} is already marked recorded.".format(marked_field)
        else:
            self.message = "Field {} in entry titled {} is already marked recorded.".format(
                marked_field, entry["Title"]
            )
        self.message += " Check if there are duplicates."
        self.args = (self.message,)



import logging
import re
from collections import defaultdict

import jieba
from pypinyin import lazy_pinyin as pinyin

from cnki2bibtex.entry import Entry, NOT_FOUND
from cnki2bibtex.misc.check import check_entry_has_valid_fields, check_bib_entry_has_id
from cnki2bibtex.misc.configure import get_id_format


class BibTeXEntry(Entry):

    rules = []

    @classmethod
    def cnki_entry_is_this_bib_entry_type(cls, cnki_entry):
        if not cls.rules or not isinstance(cls.rules, list):
            raise RulesNotValidException(cls.rules)
        for rule in cls.rules:
            if cls.__cnki_entry_match_this_rule(cnki_entry, rule):
                return True
        return False

    @staticmethod
    def __cnki_entry_match_this_rule(cnki_entry, rule):
        for key, target in rule.items():
            if cnki_entry[key] != target:
                return False
        return True

    def __init__(self, cnki_entry=None):
        self.fields = defaultdict(lambda: None)
        self.id = None
        if cnki_entry:
            self.generate_fields(cnki_entry)
            self.generate_id(cnki_entry)

    def generate_id(self, cnki_entry):
        id_format = get_id_format()
        if id_format == "title":
            self.generate_id_in_title_format(cnki_entry)
        elif id_format == "nameyear":
            self.generate_id_in_name_year_format(cnki_entry)

    def generate_id_in_name_year_format(self, cnki_entry):
        name = cnki_entry["Author"].split(";")[0].split(",")[0].split("，")[0]
        name = name.replace(" ", "").replace(u"\u3000", "")
        year = cnki_entry["Year"]
        if self.__is_full_english(name):
            self.id = name + year
        else:
            self.id = "".join([i.title() for i in pinyin(name)]) + year

    def generate_id_in_title_format(self, cnki_entry):
        title = cnki_entry["Title"]
        title = re.sub(r"[0-9]", "", title)
        title = re.sub(r"[_,;]", "", title)
        if self.__is_full_english(title):
            title_words = title.strip().split(" ")
            self.id = "".join(title_words[0 : min(len(title_words), 4)])
        else:
            jieba.setLogLevel(logging.INFO)
            title = title.replace(" ", "").replace(u"\u3000", "")
            title_words = list(jieba.cut(title))
            string_to_convert = "".join(title_words[0 : min(len(title_words), 3)])
            self.id = "".join(pinyin(string_to_convert))

    def __is_full_english(self, string):
        return not re.search(u"[\u4e00-\u9fa5]", string)

    def generate_fields(self, cnki_entry):
        self.generate_required_fields_and_mark_recorded(cnki_entry)
        self.generate_optional_fields(cnki_entry)
        self.fix_entry_differences()

    def fix_entry_differences(self):
        if "author" in self:
            self["author"] = self["author"].strip(";").replace(";;", " and ").replace(";", " and ")
            if not self.__is_full_english(self["author"]):
                self["author"] = self["author"].replace(",", " and ").replace("，", " and ")
        for field_name, field_content in self.items():
            self[field_name] = field_content.replace(r"&", r"\&").replace(r"_", r"\_")

    def generate_required_fields_and_mark_recorded(self, cnki_entry):
        raise NotImplementedError

    def generate_optional_fields(self, cnki_entry):
        for field_name in cnki_entry:
            if cnki_entry.field_is_not_recorded(field_name) and field_name not in ["Reference Type", NOT_FOUND]:
                self.record_one_optional_field(cnki_entry, field_name)

    def record_one_optional_field(self, cnki_entry, field_name):
        save_field_names = field_name.split("/")
        NOT_SET_LOWER_FIELD_NAME_LIST = ["ISBN", "ISSN"]
        for save_field_name in save_field_names:
            save_field_name = "".join(save_field_name.strip().split(" ")).replace(",", "")
            if save_field_name not in NOT_SET_LOWER_FIELD_NAME_LIST:
                self.fields[save_field_name.lower()] = cnki_entry[field_name]
            else:
                self.fields[save_field_name] = cnki_entry[field_name]
        cnki_entry.mark_fields_are_recorded(field_name)

    @check_entry_has_valid_fields
    @check_bib_entry_has_id
    def to_bib_file_string(self):
        string = ""
        string += r"@"
        string += self.__class__.__name__
        string += r"{"
        string += self.id
        string += ",\n"
        for key, value in self.items():
            string += "\t{key} = {{{value}}},\n".format(key=key, value=value)
        string += r"}"
        string += "\n\n"
        return string


class Article(BibTeXEntry):

    rules = [{"Reference Type": "Journal Article"}]

    def generate_required_fields_and_mark_recorded(self, cnki_entry):
        self["author"] = cnki_entry["Author"]
        self["title"] = cnki_entry["Title"]
        self["journal"] = cnki_entry["Journal"]
        self["year"] = cnki_entry["Year"]
        cnki_entry.mark_fields_are_recorded("Author", "Title", "Journal", "Year")


class Book(BibTeXEntry):

    rules = [{"Reference Type": "Book"}]

    def generate_required_fields_and_mark_recorded(self, cnki_entry):
        self["author"] = cnki_entry["Author"]
        self["title"] = cnki_entry["Title"]
        self["publisher"] = cnki_entry["Publisher"]
        self["year"] = cnki_entry["Date"].split("-")[0]
        cnki_entry.mark_fields_are_recorded("Author", "Title", "Publisher")

    def generate_optional_fields(self, cnki_entry):
        cnki_entry["month"] = cnki_entry["Date"].split("-")[1]
        if "图书印刷版ISBN" in cnki_entry:
            cnki_entry["ISBN"] = cnki_entry.pop("图书印刷版ISBN")
        super().generate_optional_fields(cnki_entry)


class Booklet(BibTeXEntry):
    rules = [{NOT_FOUND: NOT_FOUND}]  # There seems to be no entry on CNKI matches this type.


class InBook(BibTeXEntry):
    rules = [{NOT_FOUND: NOT_FOUND}]  # There seems to be no entry on CNKI matches this type.


class InCollection(BibTeXEntry):
    rules = [{NOT_FOUND: NOT_FOUND}]  # There seems to be no entry on CNKI matches this type.


class InProceedings(BibTeXEntry):
    rules = [{"Reference Type": "Conference Proceedings"}]

    def generate_required_fields_and_mark_recorded(self, cnki_entry):
        self["author"] = cnki_entry["Author"]
        self["title"] = cnki_entry["Title"]
        self["booktitle"] = cnki_entry["Tertiary Title"]
        self["year"] = cnki_entry["Year"]
        cnki_entry.mark_fields_are_recorded("Author", "Title", "Tertiary Title", "Year")


class Manual(BibTeXEntry):
    rules = [{NOT_FOUND: NOT_FOUND}]  # There seems to be no entry on CNKI matches this type.


class MastersThesis(BibTeXEntry):
    rules = [{"Reference Type": "Thesis", "Type of Work": "硕士"}]

    def generate_required_fields_and_mark_recorded(self, cnki_entry):
        self["author"] = cnki_entry["Author"]
        self["title"] = cnki_entry["Title"]
        self["school"] = cnki_entry["Publisher"]
        self["year"] = cnki_entry["Year"]
        cnki_entry.mark_fields_are_recorded("Author", "Title", "Publisher", "Year")


class PhdThesis(BibTeXEntry):
    rules = [{"Reference Type": "Thesis", "Type of Work": "博士"}]

    def generate_required_fields_and_mark_recorded(self, cnki_entry):
        self["author"] = cnki_entry["Author"]
        self["title"] = cnki_entry["Title"]
        self["school"] = cnki_entry["Publisher"]
        self["year"] = cnki_entry["Year"]
        cnki_entry.mark_fields_are_recorded("Author", "Title", "Publisher", "Year")


class Proceedings(BibTeXEntry):
    rules = [{NOT_FOUND: NOT_FOUND}]  # There seems to be no way to extract the proceeding citation info on CNKI.


class TechReport(BibTeXEntry):
    rules = [{NOT_FOUND: NOT_FOUND}]  # There seems to be no entry on CNKI matches this type.


class Unpublished(BibTeXEntry):
    rules = [{NOT_FOUND: NOT_FOUND}]  # There seems to be no entry on CNKI matches this type.


class Misc(BibTeXEntry):
    rules = [{NOT_FOUND: NOT_FOUND}]  # Misc is the entry that matches any case.

    def generate_required_fields_and_mark_recorded(self, cnki_entry):
        pass


class BibTeXContentStringFactory(object):
    SPECIFIED_ENTRY_TYPE_LIST = [
        Article,
        Book,
        Booklet,
        InBook,
        InCollection,
        InProceedings,
        Manual,
        MastersThesis,
        PhdThesis,
        Proceedings,
        TechReport,
        Unpublished,
    ]

    @classmethod
    def give_bib_file_content_string(cls, cnki_entries):
        full_string = ""
        for cnki_entry in cnki_entries:
            found = False

            for entry_type in cls.SPECIFIED_ENTRY_TYPE_LIST:
                if entry_type.cnki_entry_is_this_bib_entry_type(cnki_entry):
                    full_string += entry_type(cnki_entry).to_bib_file_string()
                    found = True
                    break

            if not found:
                full_string += Misc(cnki_entry).to_bib_file_string()

        return full_string


class RulesNotValidException(Exception):
    def __init__(self, rules):
        self.message = str(rules) + " is not valid. It should be a list whose elements are dictionaries."
        self.args = (self.message,)


class OneRuleIsNotValidException(Exception):
    def __init__(self, rule):
        self.message = str(rule) + " is not valid. It should be a dictionary whose values are simply String."
        self.args = (self.message,)




import os
import re


def set_id_format(id_format):
    file_path = os.path.join(os.path.expanduser('~'), r".cnki2bib.cfg")
    with open(file_path, "w", encoding="utf-8") as f:
        f.write("[settings]\nid_format = {}".format(id_format))


def get_id_format():
    file_path = os.path.join(os.path.expanduser('~'), r".cnki2bib.cfg")
    if os.path.exists(file_path):
        with open(file_path, "r", encoding="utf-8") as f:
            config_str = f.read()
        return re.search(r"id_format = (.*)", config_str).group(1)
    else:
        return "title"



import click
import pyperclip
import logging
import os

from cnki2bibtex.bibtex_entry import BibTeXContentStringFactory
from cnki2bibtex.cnki_entry import CNKIEntryFactory
from cnki2bibtex.misc.configure import set_id_format


def get_bib_file_content_string(cnki_file_content):
    cnki_entries = CNKIEntryFactory().give_all_entries(cnki_file_content)
    return BibTeXContentStringFactory.give_bib_file_content_string(cnki_entries)


def copy_to_clipboard(content):
    pyperclip.copy(content)


@click.command()
@click.argument("inputFile", type=click.Path(exists=True, dir_okay=False), required=False)
@click.option(
    "--copy/--no-copy", "-c/-nc", default=True, help="Whether or not to copy the result to clipboard. Default: True"
)
@click.option(
    "--outputDefault/--no-outputDefault",
    "-od/-nod",
    default=True,
    help="Whether or not to create a default .bib file. It has the same name as the source .net file in its directory. Or if the input source is clipboard, it will be 'out.bib' in current working directory. Default: True",
)
@click.option(
    "--outputfile", "-o", type=click.Path(exists=False, dir_okay=False), help="Create a certain output .bib file"
)
@click.option(
    "--id-format",
    "-f",
    type=click.Choice(["title", "nameyear"]),
    help="Choose the format of the ID. Pinyin of the first words in the title, or pinyin of the first author plus year",
)
@click.option(
    "--append/--no-append",
    "-a/-na",
    default=False,
    help="Whether to append at the end of output file or overwrite it. Default: False (overwrite)",
)
def launch(inputfile, copy, outputdefault, outputfile, id_format, append):
    """Converting a NoteExpress Entry .net file exported by CNKI to BibTeX .bib file.

\b
Arguments:

    INPUTFILE: Optional. The input .net file to be converted. If left empty, the contents in the clipboard will be used."""

    if id_format:
        set_id_format(id_format)
        click.echo(f"Id format set to {id_format}")

    if not copy and not outputdefault and not outputfile and not id_format:
        return

    if inputfile != None:
        if os.path.splitext(inputfile)[1] != ".net":
            logging.warning("The input file may not be a NoteExpress Entries file.")

        try:
            with open(inputfile, "r", encoding="utf-8") as f:
                cnki_file_content = f.read()
        except:
            click.echo("Failed to open the file. Please check input path.")
    else:
        click.echo("Read the NoteExpress Entry content from the clipboard.")
        cnki_file_content = pyperclip.paste()

    try:
        bib_file_string = get_bib_file_content_string(cnki_file_content)
    except Exception as e:
        click.echo("Failed to phrase the NoteExpress Entry content.")
        click.echo("Error message:\n" + str(e))

    if copy:
        try:
            copy_to_clipboard(bib_file_string)
            click.echo("BibTeX entries copied to clipboard.")
        except Exception as e:
            click.echo("Failed to copy to the clipboard.")
            click.echo("Error message:\n" + str(e))

    if outputdefault:
        try:
            target_path = (os.path.splitext(inputfile)[0] if inputfile else "out") + ".bib"
            with open(target_path, "a" if append else "w", encoding="utf-8") as f:
                f.write(bib_file_string)

            if inputfile:
                click.echo(
                    "File '{}' is created at the same directory as the source file.".format(
                        os.path.basename(target_path)
                    )
                )
            else:
                click.echo("File 'out.bib' is created at current directory.")
        except Exception as e:
            click.echo("Failed to write the default output .bib file.")
            click.echo("Error message:\n" + str(e))

    if outputfile:
        try:
            with open(outputfile, "a" if append else "w", encoding="utf-8") as f:
                f.write(bib_file_string)

            click.echo("The output file is created.")
        except Exception as e:
            click.echo("Failed to create a .bib file at the path you choose.")
            click.echo("Error message:\n" + str(e))


if __name__ == "__main__":
    # The parameter is specified by CLI argument and managed by Click
    launch()  # pylint: disable=no-value-for-parameter



from collections import defaultdict


class EntryFieldInvalidException(Exception):
    def __init__(self, class_name):
        self.message = "The fields of {} is invalid.".format(class_name)
        self.args = (self.message,)


class RequiredFieldMissingException(Exception):
    def __init__(self, field_name):
        self.message = "The required field {} is not in the target entry.".format(field_name)
        self.args = (self.message,)


class BibEntryHasNoIDException(Exception):
    def __init__(self):
        self.message = "The BibTeX entry has no ID yet."
        self.args = (self.message,)


def check_entry_has_valid_fields(func):
    def inner(self, *args, **kwargs):
        if not self.fields or not isinstance(self.fields, (dict, defaultdict)):
            raise EntryFieldInvalidException(self.__class__.__name__)
        return func(self, *args, **kwargs)

    return inner


def check_bib_entry_has_id(func):
    def inner(self, *args, **kwargs):
        if not self.id:
            raise BibEntryHasNoIDException()
        return func(self, *args, **kwargs)

    return inner
